(*CM.make("sources.cm");
(* seems necessary  even though should be loaded from sources.cm *)
CM.make("$/ml-yacc-lib.cm");*)

structure Demo = struct

  open NRC;
  open Eval;
  exception Error of string;
	    
  (* parsing functors *)
  structure NrcLrVals = NrcLrValsFun(structure Token = LrParser.Token)
  structure NrcLex = NrcLexFun(structure Tokens = 
                               NrcLrVals.Tokens)
  structure NrcParser = Join(structure ParserData = NrcLrVals.ParserData
                             structure Lex=NrcLex
                             structure LrParser=LrParser)
  
  structure Tag = SetTag;

  structure TaggedValue = TaggedValueFun(structure Tag = Tag);
  structure TaggedEval = TaggedEvalFun(structure TaggedValue = TaggedValue);
  structure TaggedValue2Html = 
  TaggedValue2HiliteFun(structure TaggedValue = TaggedValue);
	    
  structure TaggedType = TaggedTypeFun(structure Tag = Tag);
  structure TaggedTypeAnalyze = TaggedTypeAnalyzeFun(structure TaggedType = TaggedType);
  structure TaggedType2Html = TaggedType2HiliteFun(structure TaggedType = TaggedType);
	    
  structure Tagged2Table = 
  Tagged2Table(structure TaggedValue = TaggedValue 
  structure TaggedType = TaggedType);
	    
  val bgColorScript = 
      "<script> "^
      "function setBgColor (id, color) { "^
      "   if (document.layers) "^
      "document[id].bgColor = color == 'transparent' ? null : color; "^
      "  else if (document.all) "^
      "    document.all[id].style.backgroundColor = color; "^
      "  else if (document.getElementById) "^
      "    document.getElementById(id).style.backgroundColor = color;" ^
      "} "^
      "function setBgColors (id_arr, color) { "^
      "        for (var i=0;i<id_arr.length;i++) {"^
      "                setBgColor(id_arr[i],color); "^
      "        }"^
      "} "^
      "</script>";
      
 					    
  fun parse_stream stream =
      let val _ = NrcLex.UserDeclarations.pos := 0
	fun get _ = TextIO.input stream
	fun parseerror(s,p1,p2) = raise (Error(s))
	val lexer = NrcParser.Stream.streamify (NrcLex.makeLexer get)
	val (absyn, _) = NrcParser.parse(30,lexer,parseerror,())
      in absyn
      end handle LrParser.ParseError => raise LrParser.ParseError
					      
  fun parse filename = 
      let val file = TextIO.openIn filename
	  val absyn = parse_stream file
	  val () = TextIO.closeIn file
      in
	  absyn
      end handle LrParser.ParseError => raise LrParser.ParseError

  fun parse_exp filename =
      let val NRC.Exp absyn = parse filename
      in absyn
      end

  val Teams = Var.mk_var "Teams"
  val teams = eval (Var.Map.empty) (parse_exp "tests/demo_teams.nrc")
  val Players = Var.mk_var "Players"
  val players = eval (Var.Map.empty) (parse_exp "tests/demo_players.nrc")

  val db = [(Teams,teams),
	    (Players,players)
	   ];

  val teams_ty = SetTy(RecordTy([("name", StringTy),
				   ("team",StringTy),
				   ("gp",IntTy),
				   ("win",IntTy),
				   ("loss",IntTy),
				   ("tie",IntTy),
				   ("otl",IntTy)]))
  val players_ty = SetTy(RecordTy([("name",StringTy),
				     ("team",StringTy),
				     ("gp",IntTy),
				     ("g",IntTy),
				     ("a",IntTy)]))
  val ty = [(Teams,teams_ty),
	    (Players,players_ty)
	   ];
      
  fun find (ctx,k) = case Var.Map.find(ctx,k) of SOME v => v
  val ctx = List.foldr Var.Map.insert' Var.Map.empty db;
  val tyctx = List.foldr Var.Map.insert' Var.Map.empty ty;
  val tctx = Var.Map.map TaggedValue.mk_atomic_tagged_value ctx;
  val ttyctx = Var.Map.map TaggedType.mk_tagged_ty tyctx;

  val tteams = find (tctx, Teams)
  val tplayers = find (tctx, Players)
  val tteams_ty = find (ttyctx, Teams)
  val tplayers_ty = find (ttyctx, Players)

	
  val teams_html = Tagged2Table.tag2table(tteams,teams_ty)
  val players_html = Tagged2Table.tag2table(tplayers,players_ty)							   

  fun makehtml_table msg result = 
      "<table border='1' cellpadding='2' align='top'><tr><th>Teams</th><th>Players</th><th>"^msg^"</th></tr>"
      ^"<tr>"
      ^"<td>"^(Html.htmlsToString teams_html)^"</td>"
      ^"<td>"^(Html.htmlsToString players_html)^"</td>"
      ^"<td>"^(Html.htmlsToString result)^"</td>"
      ^"</tr>"
      ^"</table>"

  fun makehtmlpage msg result = 
      let 
      in
	  "<html><head><title>Test</title>"^bgColorScript^"</head><body>"
	  ^makehtml_table msg result
	  ^"</body>"
      end

  fun makehtml' exp msg = 
      let val query_result = TaggedEval.eval tctx exp
	  val ty = NRC.infer_ty (tyctx, exp)
	  val result = Tagged2Table.tag2tableref(query_result,ty)
      in
	  makehtmlpage msg result
      end

  fun makehtml_table' exp msg = 
      let val query_result = TaggedEval.eval tctx exp
	  val ty = NRC.infer_ty (tyctx, exp)
	  val result = Tagged2Table.tag2tableref(query_result,ty)
      in
	  makehtml_table msg result
      end


      
  fun makehtml queryfile msg = 
      let val str = TextIO.openOut (queryfile^".html")
	  val exp = parse_exp ("tests/"^queryfile^".nrc") 
      in 
	  TextIO.output(str,makehtml' exp msg);
	  TextIO.closeOut(str)
      end
      
  fun printhtml query msg = 
      TextIO.output(TextIO.stdOut,makehtml' query msg)
      
  fun print_table query msg = 
      TextIO.output(TextIO.stdOut,makehtml_table' query msg)
      
      
  fun main (name,[nrcfile]) =
      (
      let val _ = print ("Making "^nrcfile^"\n")
	  val _ = makehtml nrcfile ""
      in 0 
      end handle _ => 1)
    | main (name,[nrcfile,msg]) =
      (
      let val _ = print ("Making "^nrcfile^"\n")
	  val _ = makehtml nrcfile msg
      in 0 
      end handle _ => 1)
    | main (name,[]) = 
      (case parse_stream TextIO.stdIn 
	of Exp query =>
	   (print_table query "Result";
	    0)
	 | Stmt stmt => 
	   let val db_rcd = TaggedValue.Tag(Tag.empty,TaggedValue.RecordV
                            [("Teams",tteams),
			     ("Players",tplayers) ])
	       val TaggedValue.Tag(_,TaggedValue.RecordV(result)) = 
		   TaggedEval.eval_upd tctx stmt db_rcd
 	       val result_teams = Record.lookup "Teams" result 
 	       val result_players = Record.lookup "Players" result 
 	       val result_teams_html = Tagged2Table.tag2tableref(result_teams,teams_ty)
 	       val result_players_html = Tagged2Table.tag2tableref(result_players,players_ty)

	   in print ("<table border='1' cellpadding='2' align='top'><tr><th>Teams</th><th>Players</th></tr>"
		     ^"<tr>"
		     ^"<td>"^(Html.htmlsToString teams_html)^"</td>"
		     ^"<td>"^(Html.htmlsToString players_html)^"</td>"
		     ^"</tr>"
		     ^"<tr>"
		     ^"<td>"^(Html.htmlsToString result_teams_html)^"</td>"
		     ^"<td>"^(Html.htmlsToString result_players_html)^"</td>"
		     ^"</tr>"
		     ^"</table>");
	      0
	   end
      )

      
end
