structure Html =
  struct

    (* Tiny HTML dialect *)

  datatype html_element = HTML
		| HEAD
		| SCRIPT
		| BODY 
		| A 
		| P 
		| DIV 
		| SPAN
		| UL 
		| OL 
		| TABLE 
		| TD
		| TR 
		| TH
		| SUP
		| SUB
		| B
  type html_attlist = (string * string) list
  type html_tag = html_element * html_attlist
  datatype html = E of html_tag  * html list
		| S of string

  fun a_href url = (A,[("href",url)])
  fun a_name tag = (A,[("name",tag)])
  fun div_id tag = (DIV,[("id",tag)])
  fun span_id tag = (SPAN,[("id",tag)])
  fun table_border b = (TABLE,[("border",b)])
  fun td xs = E((TD,[]),xs)
  fun td' x = E((TD,[]),[x])
  fun th xs = E((TH,[]),xs)
  fun tr xs = E((TR,[]),xs)
  val sup = (SUP,[])
  val span = (SPAN,[])

  fun htmlEltToString HTML = "html"
    | htmlEltToString HEAD = "head" 
    | htmlEltToString SCRIPT = "script" 
    | htmlEltToString BODY = "body"
    | htmlEltToString A = "a"
    | htmlEltToString P = "p"
    | htmlEltToString DIV = "div"
    | htmlEltToString SPAN = "span"
    | htmlEltToString UL = "ul"
    | htmlEltToString OL = "ol"
    | htmlEltToString TABLE = "table"
    | htmlEltToString TD = "td"
    | htmlEltToString TR = "tr"
    | htmlEltToString TH = "th"
    | htmlEltToString SUP = "sup"
    | htmlEltToString SUB = "sub"
    | htmlEltToString B = "b"

  fun ifopt NONE f = []
    | ifopt (SOME x) f = [f x]

  fun attlistToString l = String.concat (List.map (fn (k,v) => " "^k^"=\""^v^"\"") l);

  fun htmlToString (S s) = s
    | htmlToString (E((elt,atts),htmls)) = 
      "<"^(htmlEltToString elt)^(attlistToString atts)^">"
      ^     htmlsToString htmls
      ^ "</"^(htmlEltToString elt)^">\n"
  and htmlsToString htmls = String.concat (List.map htmlToString htmls)




  end
