structure Tokens = Tokens
exception Error of string;

type pos = int
type svalue = Tokens.svalue
type ('a,'b) token = ('a,'b) Tokens.token
type lexresult= (svalue,pos) token

val pos = ref 0
val eof = fn () => Tokens.EOF(!pos,!pos)
val error = fn (e,l : int,_) =>
             raise Error("line " ^ (Int.toString l) ^
                               ": " ^ e ^ "\n")
fun trim s = String.substring (s,1,String.size s - 2)
%%
%header (functor NrcLexFun(structure Tokens: Nrc_TOKENS));
alpha=[A-Za-z];
digit=[0-9];
ws = [\ \t];
%%
\n       => (pos := (!pos) + 1; lex());
{ws}+    => (lex());
{digit}+ => (case Int.fromString (yytext)
	       of SOME(i) => Tokens.NUM(i,!pos,!pos)
                | NONE => error ("bad number",!pos,!pos));
"-"{digit}+ => (case Int.fromString (yytext)
	       of SOME(i) => Tokens.NUM(i,!pos,!pos)
                | NONE => error ("bad number",!pos,!pos));
"("      => (Tokens.LPAREN(!pos,!pos));
")"      => (Tokens.RPAREN(!pos,!pos));
"<"      => (Tokens.LANGLE(!pos,!pos));
">"      => (Tokens.RANGLE(!pos,!pos));
","      => (Tokens.COMMA(!pos,!pos));
"if"     => (Tokens.IF(!pos,!pos));
"then"   => (Tokens.THEN(!pos,!pos));
"else"   => (Tokens.ELSE(!pos,!pos));
"not"    => (Tokens.NOT(!pos,!pos));
"/\\"    => (Tokens.AND(!pos,!pos));
"="      => (Tokens.EQ(!pos,!pos));
"let"    => (Tokens.LET(!pos,!pos));
"in"     => (Tokens.IN(!pos,!pos));
"fst"    => (Tokens.FST(!pos,!pos));
"snd"    => (Tokens.SND(!pos,!pos));
"sum"    => (Tokens.SUM(!pos,!pos));
"empty"    => (Tokens.EMPTY(!pos,!pos));
"max"    => (Tokens.MAX(!pos,!pos));
"min"    => (Tokens.MIN(!pos,!pos));
"true"   => (Tokens.BOOL(true,!pos,!pos));
"false"  => (Tokens.BOOL(false,!pos,!pos));
"bool"   => (Tokens.BOOLTY(!pos,!pos));
"int"    => (Tokens.INTTY(!pos,!pos));
"string" => (Tokens.STRINGTY(!pos,!pos));
":"      => (Tokens.COLON(!pos,!pos));
"{"      => (Tokens.LBRACE(!pos,!pos));
"}"      => (Tokens.RBRACE(!pos,!pos));
"|"      => (Tokens.BAR(!pos,!pos));
"<-"     => (Tokens.LARROW(!pos,!pos));
"U"      => (Tokens.BIGUNION(!pos,!pos));
"u"      => (Tokens.UNION(!pos,!pos));
"\\"     => (Tokens.DIFF(!pos,!pos));
"+"      => (Tokens.PLUS(!pos,!pos));
"*"      => (Tokens.TIMES(!pos,!pos));
"["      => (Tokens.LBRACK(!pos,!pos));
"]"      => (Tokens.RBRACK(!pos,!pos));
";"      => (Tokens.SEMI(!pos,!pos));
"=>"     => (Tokens.DARROW(!pos,!pos));
"skip"   => (Tokens.SKIP(!pos,!pos));
"repl"   => (Tokens.REPL(!pos,!pos));
"add"    => (Tokens.ADD(!pos,!pos));
"drop"   => (Tokens.DROP(!pos,!pos));
"insert" => (Tokens.INSERT(!pos,!pos));
"delete" => (Tokens.DELETE(!pos,!pos));
"iter"   => (Tokens.ITER(!pos,!pos));
"usnd"   => (Tokens.USND(!pos,!pos));
"ufst"   => (Tokens.UFST(!pos,!pos));

{alpha}+ => (Tokens.ID(yytext,!pos,!pos));
"\""({alpha}|" ")*"\"" => (Tokens.STRING(trim yytext,!pos,!pos));

"."      => (Tokens.DOT(!pos,!pos));
.        => (error ((Int.toString (!pos))^":ignoring bad character "^yytext,!pos,!pos);
             lex());
