signature NRC = 
  sig
  type var = Var.var

  datatype ty = BoolTy
	      | IntTy 
	      | PairTy of ty * ty 
	      | SetTy of ty
	      | RecordTy of (string * ty) list


  datatype exp = Var of var
	       | Let of var * exp * exp
	       | Pair of exp * exp
	       | Fst of exp 
	       | Snd of exp
	       | IntC of int
	       | Plus of exp * exp
	       | Sum of exp
	       | Max of exp
	       | Min of exp
	       | Empty of exp
	       | Equal of exp * exp
	       | BoolC of bool
	       | Neg of exp
	       | And of exp * exp
	       | IfThenElse of exp * exp * exp
	       | EmptyC of ty
	       | Singleton of exp
	       | Union of exp * exp
	       | Diff of exp * exp
	       | Comp of exp * var * exp
	       | Flatten of exp
	       | StringC of string
	       (* Records *)
	       | Record of (string * exp) list
	       | Field of exp * string

    type ctx = ty Var.Map.map;

    val swap : (var * var) -> exp -> exp

    val infer_ty : ctx * exp -> ty
    val check_ty : ctx * exp * ty -> bool

  end
