structure Simplify = struct
  (* Simplifier for NRC expressions *)
  open NRC;

(* TODO: If then else simplifications for constructors *)

  fun try_opt (NONE) f = NONE
    | try_opt (SOME x) f = SOME (f x)

  fun simp_plus (IntC(i1),IntC(i2)) = IntC(i1+i2)
    | simp_plus (e1,e2) = Plus(e1,e2)
  fun simp_times (IntC(i1),IntC(i2)) = IntC(i1*i2)
    | simp_times (e1,e2) = Times(e1,e2)
  fun simp_sum (EmptyC _) = IntC 0
    | simp_sum (Singleton(e)) =  e
    | simp_sum (Union(e1,e2)) = simp_plus(simp_sum e1,simp_sum e2)
    | simp_sum e = Sum(e)
  fun simp_fst (Pair(e1,e2)) = e1
    | simp_fst e' = Fst(e')
  fun simp_snd (Pair(e1,e2)) = e2
    | simp_snd e' = Snd(e')
  fun simp_neg e = Neg(e)
  fun simp_and (e1,e2) = And(e1,e2)
  fun simp_if (BoolC b,e2,SOME e3) = if b then e2 else e3
    | simp_if (IfThenElse(e11,e12,SOME e13),e2,e3opt) = 
      IfThenElse (e11,simp_if(e12,e2,e3opt),SOME (simp_if(e13,e2,e3opt)))
    | simp_if (e1,e2,e3) = IfThenElse(e1,e2,e3)
  fun simp_union (EmptyC _,e2) = e2
    | simp_union (e1,EmptyC _) = e1
    | simp_union (IfThenElse(e11,e12,e13opt),e2) = 
      IfThenElse(e11,simp_union(e12,e2),try_opt e13opt (fn x => simp_union(x,e2)))
    | simp_union (e1,IfThenElse(e21,e22,e23opt)) = 
      IfThenElse(e21,simp_union(e1,e21),try_opt e23opt (fn x => simp_union(e1,x)))
    | simp_union (e1,e2) = Union(e1,e2)
  fun simp_comp (e1,x,EmptyC ty) = EmptyC ty
    | simp_comp (e1,x,Singleton(e)) = Let(x,e,Singleton(e1))
    | simp_comp (e1,x,Union(e21,e22)) = 
      Union(simp_comp(e1,x,e21),simp_comp(e1,x,e22))
    | simp_comp (e1,x,Comp(e2,y,e3)) = simp_comp (Let(x,e2,e1),y,e3)
    | simp_comp (e1,x,Flatten(e2)) = 
      let val y = Var.fresh_anonymous_var() 
      in 
	  Flatten(simp_comp(Comp(e1,x,Var y),y,e2))
      end
    | simp_comp (IfThenElse(e11,e12,e13opt),x,e2) = 
      IfThenElse(e11,simp_comp(e12,x,e2),
		 try_opt e13opt (fn e13 => simp_comp(e13,x,e2)))
    | simp_comp (e1,x,e2) = Comp(e1,x,e2)
  fun simp_flatten (EmptyC (SetTy(ty))) = EmptyC(ty)
    | simp_flatten (Singleton(e)) = e
    | simp_flatten (Union(e1,e2)) = Union(simp_flatten(e1),simp_flatten(e2))
    | simp_flatten (IfThenElse(e1,e2,e3opt)) = 
      IfThenElse(e1,simp_flatten(e2),try_opt e3opt (fn e3 => simp_flatten e3))
    | simp_flatten (e) = Flatten(e)


  fun simp_plus (e1,e2) = Plus(e1,e2)
  fun simp_times (e1,e2) = Times(e1,e2)
  fun simp_sum e = Sum(e)
  fun simp_fst e' = Fst(e')
  fun simp_snd e' = Snd(e')
  fun simp_neg e = Neg(e)
  fun simp_and (e1,e2) = And(e1,e2)
  fun simp_if (e1,e2,e3) = IfThenElse(e1,e2,e3)
  fun simp_union (e1,e2) = Union(e1,e2)
  fun simp_comp (e1,x,e2) = Comp(e1,x,e2)
  fun simp_flatten (e) = Flatten(e)

  fun simplify exp = 
      let fun s (Let (v,e1,e2)) = Let(v,s e1, s e2)
	    | s (Pair(e1,e2)) = Pair(s e1, s e2)
	    | s (Fst(e)) = simp_fst (s e)
	    | s (Snd(e)) = simp_snd (s e)
	    | s (Plus(e1,e2)) = simp_plus (s e1, s e2)
	    | s (Times(e1,e2)) = simp_times (s e1, s e2)
	    | s (Sum(e)) = simp_sum (s e)
	    | s (Neg e) = simp_neg(s e)
	    | s (And (e1,e2)) = simp_and(s e1, s e2)
	    | s (IfThenElse(e1,e2,NONE)) = simp_if (s e1, s e2, NONE)
	    | s (IfThenElse(e1,e2,SOME e3)) = simp_if (s e1, s e2, SOME (s e3))
	    | s (Singleton e) = Singleton (s e)
	    | s (Union (e1,e2)) = simp_union(s e1,s e2)
	    | s (Flatten(e)) = simp_flatten(s e)
	    | s (Comp (e1,x,e2)) = simp_comp(s e1, x, s e2)
	    | s e  = e
      in s exp
      end
end
		     
