functor TaggedTypeFun(structure Tag : TAG) : TAGGED_TYPE = 
struct
  structure Tag = Tag;
  open Tag;
  open NRC;  
  open Record;

  datatype raw_type = IntTy
                    | BoolTy
		    | StringTy
                    | PairTy of tag_type * tag_type
		    | SetTy of tag_type
		    | RecordTy of tag_type record
       and tag_type = Tag of {tag:tag,con: raw_type}

  fun tag(rv) = Tag{tag=Tag.fresh_anonymous_tag(),con=rv}
  fun notag(rv) = Tag{tag=Tag.empty,con=rv}
  fun addtag(Tag{tag=t0,con=c},t) = Tag{tag=Tag.union(t0,t),con=c}

  fun mk_empty_ty (NRC.BoolTy) = notag(BoolTy)
    | mk_empty_ty (NRC.IntTy)   = notag(IntTy)
    | mk_empty_ty (NRC.StringTy)   = notag(StringTy)
    | mk_empty_ty (NRC.PairTy(ty1,ty2)) = 
      notag(PairTy(mk_empty_ty(ty1),mk_empty_ty(ty2)))
    | mk_empty_ty (NRC.SetTy(ty)) = 
      notag(SetTy(mk_empty_ty ty))  
    | mk_empty_ty (NRC.RecordTy(rty)) = 
      notag(RecordTy(Record.map mk_empty_ty rty))

fun mk_tagged_ty (NRC.BoolTy) = tag(BoolTy)
    | mk_tagged_ty (NRC.IntTy)   = tag(IntTy)
    | mk_tagged_ty (NRC.StringTy)   = tag(StringTy)
    | mk_tagged_ty (NRC.PairTy(ty1,ty2)) = 
      tag(PairTy(mk_tagged_ty(ty1),mk_tagged_ty(ty2)))
    | mk_tagged_ty (NRC.SetTy(ty)) = 
      tag(SetTy(mk_tagged_ty ty))
    | mk_tagged_ty (NRC.RecordTy(rty)) = 
      tag(RecordTy(Record.map mk_tagged_ty rty))
			     
  fun erase (Tag{tag=_,con=IntTy}) = NRC.IntTy
    | erase (Tag{tag=_,con=BoolTy}) = NRC.BoolTy
    | erase (Tag{tag=_,con=StringTy}) = NRC.StringTy
    | erase (Tag{tag=_,con=PairTy(ty1,ty2)}) = NRC.PairTy(erase(ty1),erase(ty2))
    | erase (Tag{tag=_,con=SetTy(ty)}) = NRC.SetTy(erase ty)
    | erase (Tag{tag=_,con=RecordTy(rty)}) = NRC.RecordTy(Record.map erase rty)

  fun equiv (tty1,tty2) = NRC.equiv_ty (erase(tty1),erase(tty2))

  fun merge (Tag{tag=t1,con=c1}, Tag{tag=t2,con=c2}) = 
      Tag{tag=Tag.union(t1,t2),con=merge_raw (c1,c2)}
  and merge_raw (IntTy,IntTy) = IntTy
    | merge_raw (BoolTy,BoolTy) = BoolTy
    | merge_raw (StringTy,StringTy) = StringTy
    | merge_raw (PairTy(t1,t2),PairTy(t1',t2')) = 
      PairTy(merge(t1,t1'),merge(t2,t2'))
    | merge_raw (SetTy(t),SetTy(t')) = 
      SetTy(merge(t,t'))
    | merge_raw (RecordTy(rt),RecordTy(rt')) = 
      RecordTy(Record.merge merge (rt,rt'))

  fun prov (Tag{tag=t,con=BoolTy}) = t
    | prov (Tag{tag=t,con=IntTy}) = t
    | prov (Tag{tag=t,con=StringTy}) = t
    | prov (Tag{tag=t,con=PairTy(ty1,ty2)}) = 
      Tag.union(t,Tag.union(prov(ty1),prov(ty2)))
    | prov (Tag{tag=t,con=SetTy(ty)}) = 
      Tag.union (t, prov ty)
   | prov (Tag{tag=t,con=RecordTy(rty)}) = 
      List.foldl (fn ((x,ty),u') => union(prov ty,u')) t rty

  and raw_toString (IntTy) = "int"
    | raw_toString (BoolTy) = "bool"
    | raw_toString (StringTy) = "string"
    | raw_toString (PairTy (ty1,ty2)) = 
      "("^(toString ty1)^","^(toString ty2)^")"
    | raw_toString (SetTy ty) = 
      "{"^(toString ty)^"}"
    | raw_toString (RecordTy rty) = 
      (Record.toString toString rty)
  and toString (Tag{tag=t,con=c}) = 
	  raw_toString(c)^"["^(Tag.toString t)^"]"

  fun tag_simplify t (Tag{tag=t0,con=c}) = 
      Tag{tag=Tag.difference(t0,t),
	  con=raw_simplify (Tag.union (t,t0)) c}
  and raw_simplify t (PairTy (ty1,ty2)) = PairTy(tag_simplify t ty1,tag_simplify t ty2)
    | raw_simplify t (SetTy (ty)) = SetTy(tag_simplify t ty)
    | raw_simplify t (RecordTy (rty)) = RecordTy(Record.map (tag_simplify t) rty)
    | raw_simplify t (x) = x
  fun simplify tty = tag_simplify Tag.empty tty


end

