functor TaggedTypeAnalyzeFun(structure TaggedType : TAGGED_TYPE) = struct
  open NRC;
  open TaggedType;
  open TaggedType.Tag;

  type ctx =  tag_type Var.Map.map;

  fun analyze_prov (ctx,exp) =  
      let fun analyze (Var(v)) = 
              (case Var.Map.find(ctx,v)
		of SOME tty => tty)
            | analyze (Let(v,e1,e2)) = 
              let val tty = analyze e1
              in
		  analyze_prov (Var.Map.insert(ctx,v,tty), e2)
              end
	    | analyze (Pair(e1,e2)) = notag(PairTy(analyze e1, analyze e2))
            | analyze (Fst(e)) =
	      (case analyze e 
		of Tag{tag=t,con=PairTy(t1,_)} => addtag(t1,t))
            | analyze (Snd(e)) = 
	      (case analyze e 
		of Tag{tag=t,con=PairTy(_,t2)} => addtag(t2,t))
            | analyze (IntC(i)) = notag(IntTy)
            | analyze (Plus(e1,e2)) = 
              (case (analyze e1, analyze e2) 
		of (Tag{tag=t1,con=IntTy},
		    Tag{tag=t2,con=IntTy}) => 
		   Tag{tag=Tag.union(t1,t2),con=IntTy})
            | analyze (Times(e1,e2)) = 
              (case (analyze e1, analyze e2) 
		of (Tag{tag=t1,con=IntTy},
		    Tag{tag=t2,con=IntTy}) => 
		   Tag{tag=Tag.union(t1,t2),con=IntTy})
            | analyze (Sum(e)) = 
              (case (analyze e) 
		of Tag{tag=t,con=SetTy(Tag{tag=t0,con=IntTy})} => 
		   Tag{tag=Tag.union(t,t0),con=IntTy})
            | analyze (Min(e)) = 
              (case (analyze e) 
		of Tag{tag=t,con=SetTy(Tag{tag=t0,con=IntTy})} => 
		   Tag{tag=Tag.union(t,t0),con=IntTy})
            | analyze (Max(e)) = 
              (case (analyze e) 
		of Tag{tag=t,con=SetTy(Tag{tag=t0,con=IntTy})} => 
		   Tag{tag=Tag.union(t,t0),con=IntTy})
            | analyze (Empty(e)) = 
              (case (analyze e) 
		of Tag{tag=t,con=SetTy(_)} => 
		   Tag{tag=t,con=BoolTy})
            | analyze (Equal(e1,e2)) = 
	      let val tty1 = analyze(e1) 
		  val tty2 = analyze(e1)  
	      in
		  if equiv(tty1,tty2)
		  then Tag{tag=Tag.union(prov tty1, prov tty2),con=BoolTy}
		  else raise Match
	      end
            | analyze (BoolC(b)) = notag(BoolTy)
            | analyze (Neg(e)) = 
	      (case analyze e 
		of Tag{tag=t,con=BoolTy} => Tag{tag=t,con=BoolTy})
            | analyze (And(e1,e2)) = 
	      (case (analyze e1, analyze e2) 
		of (Tag{tag=t1,con=BoolTy},
		    Tag{tag=t2,con=BoolTy}) =>
		   Tag{tag=Tag.union(t1,t2),con=BoolTy})
            | analyze (IfThenElse(e0,e1,SOME e2)) = 
	      (case analyze e0 
		of Tag{tag=t0,con=BoolTy} => 
		   let val tty1 = analyze e1
		       val tty2 = analyze e2 
		   in if equiv(tty1,tty2) 
		      then addtag(merge(tty1,tty2) ,t0)
		      else raise Match
		   end)
            | analyze (IfThenElse(e0,e1,NONE)) = 
	      (case analyze e0 
		of Tag{tag=t0,con=BoolTy} => 
		   let val tty1 = analyze e1
		   in case tty1 of 
			  Tag{tag=_,con=SetTy(_)} => addtag(tty1,t0)
			| _ => raise Match
		   end)
            | analyze (EmptyC ty) = mk_empty_ty(NRC.SetTy  ty)
            | analyze (Singleton(e)) = notag( SetTy(analyze e))
            | analyze (Union(e1,e2)) = 
              (case (analyze e1, analyze e2) 
		of (Tag{tag=t1,con=SetTy(ty1)},
		    Tag{tag=t2,con=SetTy(ty2)}) => 
		   if equiv(ty1,ty2)
		   then Tag{tag=Tag.union(t1,t2),con=SetTy(merge(ty1,ty2))}
		   else raise Match)
            | analyze (Diff(e1,e2)) =             
              (case (analyze e1, analyze e2) 
		of (ty1 as Tag{tag=t1,con=SetTy(ty1')} ,
		    ty2 as Tag{tag=t2,con=SetTy(ty2')}) => 
		   if equiv(ty1',ty2')
		   then Tag{tag=Tag.union(prov ty1, prov ty2),
			    con=SetTy(ty1)}
		   else raise Match)
            | analyze (Comp(e1,v,e2)) = 
              (case analyze e2 
		of Tag{tag=t,con=SetTy(ty)} => 
		   Tag{tag=t,
		       con=SetTy(analyze_prov (Var.Map.insert(ctx,v,ty),e1))})
            | analyze (Flatten(e)) = 
              (case analyze e
		of Tag{tag=t,con=SetTy(Tag{tag=t0,con=SetTy(ty)})} => 
		   Tag{tag=Tag.union(t,t0),con=SetTy(ty)})
	    | analyze (StringC(s)) = notag StringTy
	    | analyze (Record(field_exps)) = 
	      notag(RecordTy(Record.map analyze field_exps))
            | analyze (Field(e,lab)) =
	      (case analyze e 
		of Tag{tag=t,con=RecordTy(rty)} => 
		   addtag(Record.lookup lab rty,t))

      in analyze exp 
      end
end

