functor TaggedValueFun(structure Tag : TAG) : TAGGED_VALUE = struct
  structure Tag = Tag;
  open Tag;
  
  datatype raw_value = BoolV of bool
		     | IntV of int
		     | StringV of string
		     | PairV of tag_value * tag_value
		     | SetV of tag_value list
		     | RecordV of tag_value Record.record
       and tag_value = Tag of Tag.tag * raw_value

  fun tag(rv) = Tag(fresh_anonymous_tag(),rv)
  fun retag(Tag(_,rv)) = Tag(fresh_anonymous_tag(),rv)
  fun notag(rv) = Tag(empty,rv)
  fun addtag(Tag(t0,c),t) = Tag (Tag.union(t0,t),c)

  fun tagOf(Tag (t,v)) = t

  fun mk_tagged_value (Value.BoolV(b)) = tag(BoolV(b))
    | mk_tagged_value (Value.IntV(i))  = tag(IntV(i))
    | mk_tagged_value (Value.StringV(s))  = tag(StringV(s))
    | mk_tagged_value (Value.PairV(v1,v2)) = 
      tag(PairV(mk_tagged_value(v1),mk_tagged_value(v2)))
    | mk_tagged_value (Value.SetV(l)) = 
      tag(SetV(List.map mk_tagged_value l))
    | mk_tagged_value (Value.RecordV(r)) = 
      tag(RecordV(Record.map mk_tagged_value r))

  fun mk_atomic_tagged_value (Value.BoolV(b)) = tag(BoolV(b))
    | mk_atomic_tagged_value (Value.IntV(i))  = tag(IntV(i))
    | mk_atomic_tagged_value (Value.StringV(s))  = tag(StringV(s))
    | mk_atomic_tagged_value (Value.PairV(v1,v2)) = 
      notag(PairV(mk_atomic_tagged_value(v1),mk_atomic_tagged_value(v2)))
    | mk_atomic_tagged_value (Value.SetV(l)) = 
      tag(SetV(List.map mk_atomic_tagged_value l))
    | mk_atomic_tagged_value (Value.RecordV(r)) = 
      notag(RecordV(Record.map mk_atomic_tagged_value r))

  fun mk_row_tagged_value (Value.BoolV(b)) = notag(BoolV(b))
    | mk_row_tagged_value (Value.IntV(i))  = notag(IntV(i))
    | mk_row_tagged_value (Value.StringV(s))  = notag(StringV(s))
    | mk_row_tagged_value (Value.PairV(v1,v2)) = 
      notag(PairV(mk_row_tagged_value(v1),mk_row_tagged_value(v2)))
    | mk_row_tagged_value (Value.RecordV(r)) = 
      notag(RecordV(Record.map mk_row_tagged_value r))
    | mk_row_tagged_value (Value.SetV(l)) = 
      notag(SetV(List.map (fn x => retag(mk_row_tagged_value x)) l))
	    

  datatype dir = Left | Right | In of int
  type path = dir list

  fun change_at(_,[],v') = v'
    | change_at(Tag(t,PairV(v1,v2)),Left::p,v') = 
      Tag(t,PairV(change_at(v1,p,v'),v2))
    | change_at(Tag(t,PairV(v1,v2)),Right::p,v') = 
      Tag(t,PairV(v1,change_at(v2,p,v')))
    | change_at(Tag(t,SetV(vs)),In(i)::p,v') = 
      Tag(t,SetV(change_at_set(vs,i,p,v')))
  and change_at_set(v::vs,0,p,v') = 
      change_at(v,p,v')::vs
    | change_at_set(v::vs,i,p,v') = 
      v::change_at_set(vs,i-1,p,v')

      
  fun erase (Tag(_,IntV(i))) = Value.IntV(i)
    | erase (Tag(_,BoolV(b))) = Value.BoolV(b)
    | erase (Tag(_,StringV(s))) = Value.StringV(s)
    | erase (Tag(_,PairV(v1,v2))) = Value.PairV(erase(v1),erase(v2))
    | erase (Tag(_,SetV(vs))) = Value.SetV(List.map erase vs)
    | erase (Tag(_,RecordV(vs))) = Value.RecordV(Record.map erase vs)

  fun raw_eq (IntV i1,IntV i2) = i1 = i2
    | raw_eq (BoolV b1,BoolV b2) = b1 = b2
    | raw_eq (StringV s1,StringV s2) = s1 = s2
    | raw_eq (PairV(v1,v2),PairV(v1',v2')) = 
      eq(v1,v1') andalso eq(v2,v2')
    | raw_eq (SetV(vs),SetV(vs')) = tag_set_eq(vs,vs')
    | raw_eq (RecordV(vs),RecordV(vs')) = Record.equiv eq (vs,vs')
    (* TODO: Set equality not list equality *)
  and tag_set_eq ([],[]) = true
    | tag_set_eq (v::vs,v'::vs') = 
      eq(v,v') andalso tag_set_eq(vs,vs')
    | tag_set_eq (_,_) = false

  and eq (Tag(_,c),Tag(_,c')) = raw_eq(c,c')
      
  fun raw_prov (BoolV(b)) = Tag.empty
    | raw_prov (IntV(i)) = Tag.empty
    | raw_prov (StringV(s)) = Tag.empty
    | raw_prov (PairV(v1,v2)) = 
      union(prov(v1),prov(v2))
    | raw_prov (SetV(vs)) = 
      List.foldl union Tag.empty (List.map prov vs)
    | raw_prov (RecordV(vs)) = 
      List.foldl union Tag.empty (List.map (fn (_,x) => prov x) vs)
  and prov (Tag(t,v)) = union (t, raw_prov v)

  fun raw_toString (IntV i) = Int.toString i
    | raw_toString (BoolV b) = Bool.toString b
    | raw_toString (StringV s) = "\""^s^"\""
    | raw_toString (PairV (v1,v2)) = 
      "("^(toString v1)^","^(toString v2)^")"
    | raw_toString (SetV vs) = 
      "{"^(tag_values_toString vs)^"}"
    | raw_toString (RecordV vs) = 
      "{"^(Record.toString toString vs)^"}"

  and tag_values_toString [] = ""
    | tag_values_toString [v] = toString v
    | tag_values_toString (v::vs) = 
      (toString v)^","^(tag_values_toString vs)

  and toString (Tag(t,c)) = 
      raw_toString(c)^"["^(Tag.toString t)^"]"
      


  fun tag_simplify t (Tag(t0,r)) = 
      Tag(difference(t0,t),raw_simplify (union (t,t0)) r)
  and raw_simplify t (PairV (v1,v2)) =PairV(tag_simplify t v1,tag_simplify t v2)
    | raw_simplify t (SetV (vs)) = SetV(List.map (tag_simplify t) vs)
    | raw_simplify t (RecordV (vs)) = RecordV(Record.map (tag_simplify t) vs)
    | raw_simplify t (x) = x
  fun simplify tv = tag_simplify Tag.empty tv

(*  fun tag_simplify (Tag(t0,r)) = 
      Tag(difference(t0,raw_prov(r)),raw_simplify r)
  and raw_simplify (PairV (v1,v2)) =PairV(tag_simplify v1,tag_simplify v2)
    | raw_simplify (SetV (vs)) = SetV(List.map tag_simplify  vs)
    | raw_simplify (RecordV (vs)) = RecordV(Record.map tag_simplify  vs)
    | raw_simplify (x) = x
  fun simplify tv = tag_simplify tv*)


(* Traverses value.  Retags each part and builds mapping from new tags to old ones *)


  fun tag_ext m (Tag (t,r)) = 
      let val (r',m') = raw_ext m r 
	  val t' = fresh_anonymous_tag()
      in (Tag(t',r'), (t',t)::m')
      end
  and raw_ext m (IntV i) = (IntV i, [])
    | raw_ext m (StringV s) = (StringV s, [])
    | raw_ext m (BoolV b) = (BoolV b, [])
    | raw_ext m (PairV (v1,v2)) = 
      let val (v1',m1) = tag_ext m v1
	  val (v2',m2) = tag_ext m1 v2
      in (PairV(v1',v2'), m2)
      end
    | raw_ext m (RecordV (fvs)) =
      let fun extract_fields ((f,v),(fvs,m)) = 
	      let val (v',m') = tag_ext m v 
	      in ((f,v)::fvs,m')
	      end
	  val (fvs',m') = List.foldr extract_fields ([],m) fvs
      in 
	  (RecordV(fvs'),m')
      end
    | raw_ext m (SetV (vs)) =
      let fun extract_elts (v,(vs,m)) = 
	      let val (v',m') = tag_ext m v 
	      in (v::vs,m')
	      end
	  val (vs',m') = List.foldr extract_elts ([],m) vs
      in 
	  (SetV(vs'),m')
      end

  fun tag_extract tv = tag_ext [] tv

end



