structure Value = struct
  open Record
     
  (* TODO: Use a smarter implementation of (multi)sets *)
  datatype value = BoolV of bool
		 | IntV of int
		 | StringV of string
		 | PairV of value * value
		 | SetV of value list
		 | RecordV of value record

  fun eq (IntV i1,IntV i2) = i1 = i2
    | eq (BoolV b1,BoolV b2) = b1 = b2
    | eq (StringV s1,StringV s2) = s1 = s2
    | eq (PairV(v1,v2),PairV(v1',v2')) = 
      eq(v1,v1') andalso eq(v2,v2')
    (* TODO: Set equality not list equality *)
    | eq (SetV(vs),SetV(vs')) = set_eq(vs,vs')
    | eq (RecordV(r),RecordV(r')) = record_eq (r,r')
  and set_eq ([],[]) = true
    | set_eq (v::vs,v'::vs') = eq(v,v') andalso set_eq(vs,vs')
    | set_eq (_,_) = false
  and record_eq (r,r') = Record.equiv eq (r, r')

  fun hom z s h [] = z
    | hom z s h [x] = h x
    | hom z s h (x::xs) = s (h x,hom z s h xs)

  fun toString (IntV i) = Int.toString i
    | toString (BoolV b) = Bool.toString b
    | toString (StringV s) = "\""^s^"\""
    | toString (PairV (v1,v2)) = "("^(toString v1)^","^(toString v2)^")"
    | toString (SetV vs) = 
      "{"^(hom "" (fn (x,y) => x^","^y) toString vs)^"}"
    | toString (RecordV vs) = 
      "{"^(hom "" (fn (x,y) => x^","^y) (fn (f,x) => f ^ ":" ^ toString x) vs)^"}"
  fun latex_value (IntV i) = Int.toString i
    | latex_value (BoolV b) = Bool.toString b
    | latex_value (StringV s) = "\""^s^"\""
    | latex_value (PairV (v1,v2)) = "("^(latex_value v1)^","^(latex_value v2)^")"
    | latex_value (SetV vs) = 
      "\\{"^(hom "" (fn (x,y) => x^","^y) latex_value vs)^"\\}"
    | latex_value (RecordV vs) = 
      "\\langle"^(hom "" (fn (x,y) => x^","^y) (fn (f,x) => f ^ ":" ^ latex_value x) vs)^"\\rangle"

end
